"""
This module will execute the dt_misc package demonstrations, which include:

- Logging demo
- OS demo
- Project helper demo
- Helper demo

To Run:
    ``poetry run python -m dt_tools.cli.demos.dt_misc_demo``

"""
from loguru import logger as LOGGER

import dt_tools.cli.demos.dt_misc_helper_demo as misc_helper_demo
import dt_tools.cli.demos.dt_logging_demo as logging_demo
import dt_tools.cli.demos.dt_os_demo as os_helper_demo
import dt_tools.cli.demos.dt_project_helper_demo as project_helper_demo
import dt_tools.cli.demos.dt_sysinfo_demo as sysinfo_demo
import dt_tools.logger.logging_helper as lh
from dt_tools.os.os_helper import OSHelper
from dt_tools.os.project_helper import ProjectHelper

def demo():
    OSHelper.enable_ctrl_c_handler()
    DEMOS = {
        "Logging demo": logging_demo,
        "OS demo": os_helper_demo,
        "System Info": sysinfo_demo,
        "Misc Helper demo": misc_helper_demo,
        "Project Helper demo": project_helper_demo,
    }
    l_handle = lh.configure_logger(log_level="INFO", brightness=False)
    LOGGER.info('='*80)
    version = f'v{ProjectHelper.determine_version("dt-foundation")}'
    LOGGER.info(f'dt_misc_demo {version}', 80)
    LOGGER.info('='*80)
    LOGGER.info('')
    for name, demo_module in DEMOS.items():
        if input(f'Run {name} (y/n)? ').lower() == 'y':
            if demo_module == logging_demo:
                # Logging demo sets up and removes it's own loggers 
                LOGGER.remove(l_handle)
            demo_module.demo()  
            if demo_module == logging_demo:
                l_handle = lh.configure_logger(log_level="INFO", brightness=False)
            LOGGER.info('') 
                                                      
    LOGGER.success("That's all folks!!")

if __name__ == '__main__':
    demo()


"""
This module will execute the dt_misc helper demo.

Features ObjectHelper and StringHelper classes.

"""
from loguru import logger as LOGGER

import dt_tools.logger.logging_helper as lh
from dt_tools.misc.helpers import ObjectHelper, StringHelper

class _Stats():
    def __init__(self, age: int, height: str, weight:int):
        self.age = age
        self.height = height
        self.weight = weight

class _Person():
    def __init__(self, name: str, statistics: _Stats, occupation: str):
        self.name: str = name
        self.stats: _Stats = statistics
        self.occupation: str = occupation

_person_dict = {
    'name': "Alberto", "stats": {"age": 45, "height": "6'1\"", "weight": 175}, "occupation": "engineer"
    }

def demo():
    LOGGER.info('')
    LOGGER.info('-'*40)
    LOGGER.info('dt_misc_helper_demo')
    LOGGER.info('-'*40)
    LOGGER.info('')

    LOGGER.info("Object to Dictionary demo")
    LOGGER.info("-------------------------")
    LOGGER.info("class Stats():")
    LOGGER.info('    def __init__(self, age:int, height:str, weight: int):')
    LOGGER.info("        self.age = age")
    LOGGER.info("        self.height = height")
    LOGGER.info("        self.weight = weight")
    LOGGER.info("")
    LOGGER.info("class Person():")
    LOGGER.info("    def __init__(self, name: str, statistics: Stats, occupation: str):")
    LOGGER.info("        self.name = name")
    LOGGER.info("        self.stats = statistics")
    LOGGER.info("        self.occupation = occupation")
    LOGGER.info("")
    LOGGER.info("stat = Stats(20, '5\'7\"', 150)")
    LOGGER.info("person = Person('Joe', stat, 'Carpenter')")
    LOGGER.info("")
    LOGGER.warning("print(ObjectHelper.to_dict(person))")
    LOGGER.info("")
    LOGGER.info("Returns:")
    stat = _Stats(20, '5\'7"', 150)
    person = _Person('Joe', stat, 'Carpenter')
    LOGGER.success(f'  {ObjectHelper.to_dict(person)}')
    LOGGER.info("")
    input('Press Enter to continue')

    LOGGER.info("")
    LOGGER.info("Dictionary to Object demo")
    LOGGER.info('-------------------------')
    LOGGER.info("person_dict = {")
    LOGGER.info('  {"name": "Alberto", "stats": {"age": 45, "height": "6\'1\"", "weight": 175}, "occupation": "engineer"}')
    LOGGER.info("")
    LOGGER.info("person = (ObjectHelper.dict_to_obj(person_dict)")
    LOGGER.info("")
    LOGGER.warning("print(f'Person object: {person}')")
    LOGGER.warning("print(f'Name: {person.name}')")
    LOGGER.warning("print(f'Age : {person.stats.age}')")
    LOGGER.info("")
    LOGGER.info("Returns:")
    person = ObjectHelper.dict_to_obj(_person_dict)
    LOGGER.success(f"  Person object: {person}")
    LOGGER.success(f'  Name: {person.name}')
    LOGGER.success(f'  Age : {person.stats.age}')
    LOGGER.info("")
    input('Press Enter to continue')
    LOGGER.info("")
    LOGGER.info("String Helper demos")
    LOGGER.info("-------------------")
    LOGGER.info("")
    LOGGER.info("print(StringHelper.pad_l(text=' Pad Left', length=20, pad_char='*'))")
    LOGGER.success(StringHelper.pad_l(text=' Pad Left', length=20, pad_char='*'))

    LOGGER.info("print(StringHelper.pad_r(text='Pad Right ', length=20, pad_char='*'))")
    LOGGER.success(StringHelper.pad_r(text='Pad Right ', length=20, pad_char='*'))

    LOGGER.info("print(StringHelper.center(text=' center ', length=20, pad_char='*'))")
    LOGGER.success(StringHelper.center(text=' center ', length=20, pad_char='*'))

    LOGGER.info('')
    LOGGER.info('Demo commplete.')            
    input('\nPress Enter to continue... ')

if __name__ == "__main__":
    lh.configure_logger(log_format=lh.DEFAULT_CONSOLE_LOGFMT, log_level="INFO", brightness=False)
    demo()



"""
This module demonstrates the ProjectHelper class features.

- Determine distribion version.
- Determine version based on file/directory name within the call stack.
- List environment installed python packages

"""
from loguru import logger as LOGGER

import dt_tools.logger.logging_helper as lh
from dt_tools.os.project_helper import ProjectHelper


def demo():
    LOGGER.info('-'*80)
    LOGGER.info('dt_misc_project_helper_demo')
    LOGGER.info('-'*80)
    LOGGER.info('')

    LOGGER.info('Determine versions')
    LOGGER.info('------------------')
    LOGGER.info(f'- Distribution dt-foundation : {ProjectHelper.determine_version("dt-foundation", identify_src=True)}')
    LOGGER.success('  ProjectHelper.determine_version("dt-foundation", identify_src=True)')
    LOGGER.info('')
    LOGGER.info(f'- File: project_helper : {ProjectHelper.determine_version("project_helper", identify_src=True)}')
    LOGGER.success('  ProjectHelper.determine_version("project_helper", identify_src=True)')
    input('\nPress Enter to continue... ')

    LOGGER.info('')
    LOGGER.info('Installed Distribution Packages')
    LOGGER.info('-------------------------------')
    LOGGER.info('')
    LOGGER.success('  ProjectHelper.installed_distribution_packages().items()')
    LOGGER.info('')
    LOGGER.info('  Package                        Version')
    LOGGER.info('  ------------------------------ ---------')
    for package, ver in ProjectHelper.installed_distribution_packages().items():
        print(f'  {package:30} {ver}')

    input('\nPress Enter to continue... ')

    LOGGER.info('')
    LOGGER.info('Installed pyproject.toml Packages')
    LOGGER.info('---------------------------------')
    LOGGER.info('')
    LOGGER.success('  ProjectHelper.installed_pyproject_toml_packages().items()')
    LOGGER.info('')
    LOGGER.info('  Package                        Version')
    LOGGER.info('  ------------------------------ ---------')
    for package, ver in ProjectHelper.installed_pyproject_toml_packages().items():
        print(f'  {package:30} {ver}')

    LOGGER.info('')
    LOGGER.info('Demo commplete.')            
    input('\nPress Enter to continue... ')

if __name__ == "__main__":
    lh.configure_logger(log_format=lh.DEFAULT_CONSOLE_LOGFMT, log_level="INFO")
    demo()



"""
This module demonstrates the logging_helper modules features.

- Console logging
- File logging
- Log colorization

"""

import datetime
import random
import time

from loguru import logger as LOGGER

import dt_tools.logger.logging_helper as lh

@lh.logger_wraps(level='INFO')
def demo_logger_wraps():
    LOGGER.info('Inside the demo_logger_wraps_function')
    time.sleep(1)

@lh.timer_wraps(level='INFO')
def demo_timer_wraps():
    LOGGER.info('Inside the demo_timer_wraps_function')
    delay = time.time() % 10
    time.sleep(delay)

def demo():
    test1_log = "./test1.log"
    test2_log = "./test2.log"
    rotation=datetime.timedelta(seconds=10)
    retention= 5

    l_handle = lh.configure_logger(log_level="TRACE")
    LOGGER.info('-'*40)
    LOGGER.info('dt_misc_logging_demo')
    LOGGER.info('-'*40)
    LOGGER.info('')
    LOGGER.info('log to console demo...')
    LOGGER.info('')

    lh.set_log_levels_brightness(True)
    LOGGER.info('')
    LOGGER.info('Log Levels BRIGHTNESS enabled')
    LOGGER.info('-----------------------------')
    lh._print_log_level_definitions()
    time.sleep(5)
    LOGGER.waitfor_complete()

    lh.set_log_levels_brightness(False)
    LOGGER.info('')
    LOGGER.info('Log Levels BRIGHTNESS disabled')
    LOGGER.info('------------------------------')
    lh._print_log_level_definitions()
    LOGGER.waitfor_complete()

    input('\nPress Enter to continue... ')

    LOGGER.info('')
    
    h_console = lh.configure_logger(log_level="TRACE", log_handle=l_handle, brightness=False)
    h_test1   = lh.configure_logger(log_target=test1_log, log_level="DEBUG")
    h_test2   = lh.configure_logger(log_target=test2_log, log_level="INFO",
                                    retention=retention, rotation=rotation)   
    LOGGER.info('Multiple logger test (console and 2 files)') 
    LOGGER.info('------------------------------------------')
    LOGGER.info('- 30 message with random log levels will be sent to the logger.')
    LOGGER.info('- Based on configuration, each message will be routed to the appropriate logger(s)')
    LOGGER.info('')
    LOGGER.info('Logger configuration:')
    LOGGER.info('  Console   : CRITICAL, ERROR, WARNING, INFO, DEBUG, TRACE')
    LOGGER.info('  Test1.log : CRITICAL, ERROR, WARNING, INFO, DEBUG')
    LOGGER.info('  Test2.log : CRITICAL, ERROR, WARNING, INFO')
    LOGGER.info('')
    LOGGER.info('NOTE: ')
    LOGGER.info('- The Console will receive ALL log messages')
    LOGGER.info(f'- The {test1_log} file will get DEBUG level and above.')
    LOGGER.info(f'- The {test2_log} file will get INFO level and above.')
    LOGGER.info(f'- The {test2_log} file is set to rotate every 10 seconds and have 5 total versions.')
    LOGGER.waitfor_complete()

    time.sleep(3)
    LOGGER.info('')
    LOGGER.trace('This TRACE message should ONLY print on Console')
    LOGGER.debug('This DEBUG message should print in test1.log and Console')
    LOGGER.info('This INFO message should print on ALL log outputs')
    print('')
    for i in range(31):
        log_level = random.choice(['TRACE','DEBUG','INFO','WARNING','ERROR','CRITICAL'])
        LOGGER.log(log_level, f'message {i:2} {log_level}')
        time.sleep(.25)
    LOGGER.waitfor_complete()

    print('Removing file handlers, resetting console to debug format.')
    LOGGER.remove(h_test1)
    LOGGER.remove(h_test2)
    # Reset console handler
    h_console = lh.configure_logger(log_level="INFO", log_format=lh.DEFAULT_DEBUG_LOGFMT, log_handle=h_console, brightness=False)
    LOGGER.info('')
    LOGGER.info('decorator: logger_wraps()')
    demo_logger_wraps()
    LOGGER.waitfor_complete()

    LOGGER.info('')
    LOGGER.info('decorator: timer_wraps()')
    demo_timer_wraps()
    LOGGER.waitfor_complete()

    h_console = lh.configure_logger(log_level="INFO", log_handle=h_console, brightness=False)
    LOGGER.info('')
    LOGGER.success('logging demo complete.')
    LOGGER.waitfor_complete()
    
    input('\nPress Enter to continue... ')
    LOGGER.remove(h_console)
    
if __name__ == "__main__":
    demo()



import tomllib
import inspect
import pathlib
from datetime import datetime as dt
from importlib.metadata import Distribution, distributions, version
from typing import Tuple, Union

from loguru import logger as LOGGER


class ProjectHelper:
    """
    Helper class for retrieving project info such as version and installed packages.

    """
    _max_depth = 4

    @staticmethod
    def _search_down_tree(filename: str, start_path: str, depth: int = _max_depth) -> Union[pathlib.Path, None]:
        """
        Search directory tree towards root for filename starting at start path.

        Limit seach to depth number of directories.

        Args:
            filename (str): target_filename.
            start_path (str): path to begin search.
            depth (int, optional): Number of directories to travers (towards root). Defaults to _max_depth.

        Returns:
            str: Full filename/path or None.
        """
        LOGGER.trace(f'  Search for {filename}')
        result_file: pathlib.Path = None
        cur_depth = 0
        traverse_path = pathlib.Path(start_path)
        pattern = f"**/{filename}"
        while result_file is None and cur_depth < depth:
            cur_depth += 1
            file_list = list(traverse_path.glob(pattern))
            LOGGER.trace(f'  - directory: {str(traverse_path)}')
            LOGGER.trace(f'    file_list: {file_list}')
            if len(file_list) == 1:
                result_file = file_list[0]
                LOGGER.trace(f'  FOUND: {result_file}')
            else:
                traverse_path = traverse_path.parent
        return result_file

    @staticmethod
    def _check_metadata(target_name: str) -> Tuple[str,str]:
        ver = None
        determined_from = None
        try:
            LOGGER.trace('Try import.metadata')
            ver = version(target_name)
            determined_from = "importlib.metadata"
        except:  # noqa: E722
            LOGGER.trace('- Not found in metadata')


        return ver, determined_from

    @staticmethod
    def _check_toml(project_name: str, calling_module: str) -> Tuple[str,str]:
        LOGGER.trace('Try pyproject.toml')
        ver = None
        LOGGER.trace(project_name)
        determined_from = None
        target_file = ProjectHelper._search_down_tree("pyproject.toml", pathlib.Path(calling_module).parent)
        # LOGGER.warning(target_file)
        if target_file is None:
            LOGGER.trace('- unable to locate pyproject.toml')
        else:
            buff = target_file.read_text(encoding='utf-8').splitlines()
            proj_name = ""
            name_list = [x for x in buff if x.startswith('name')]
            # LOGGER.warning(name_list)
            if len(name_list) >= 1:
                token = name_list[0].split('=')[1].strip()
                proj_name = token.replace('"',"").replace("'","")
                # LOGGER.warning(f'token: {token}  proj_name: {proj_name}')
            if project_name != proj_name:
                LOGGER.trace(f'- Requested project name {project_name} does not match pyproject.toml project name {proj_name}')
            else:
                ver_line = [x for x in buff if x.startswith('version')]
                LOGGER.trace(f'ver_line: {ver_line}')
                if len(ver_line) == 1:
                    ver = ver_line[0].split('=')[1].replace('"','').replace("'",'').strip()
                    determined_from = "pyproject.toml"
                LOGGER.trace('- Identified via pyproject.toml')
        return ver, determined_from
    
    @staticmethod
    def _check_call_stack(root_path: str, python_file: str) -> Tuple[str, str]:
        LOGGER.trace('Try python file')
        ver = None
        determined_from = None
        if not python_file.endswith('.py'):
            LOGGER.trace(f'  Passed file: {python_file} does not appear to be python file.')
            return ver, determined_from
        
        LOGGER.trace(f'- python file: {python_file}  root_path: {root_path}')
        file_list = list(pathlib.Path(root_path).glob(f"**/{python_file}"))
        if len(file_list) == 0:
            LOGGER.trace(f'  Unable to locate python file: {python_file}')
            return ver, determined_from
        LOGGER.trace(f'  Python file abs: {file_list[0]}')
        file_list = list(pathlib.Path(root_path).glob('**/*.py'))
        ver_date = dt(2000,1,1,0,0,0,0)
        LOGGER.trace(f'  File list: {file_list}')
        for file_nm in file_list:
            if dt.fromtimestamp(file_nm.stat().st_mtime) > ver_date:
                ver_date = dt.fromtimestamp(file_nm.stat().st_mtime)
                ver_file = file_nm
            ver_date = max(ver_date, dt.fromtimestamp(file_nm.stat().st_mtime))
        ver = f'{ver_date.year}.{ver_date.month}.{ver_date.day}'    
        determined_from = f'File date: {str(ver_date)} {ver_file}'
    
        return ver, determined_from
    
    @staticmethod
    def _get_caller() -> pathlib.Path:
        root_idx = len(inspect.stack()) - 1
        caller = pathlib.Path(inspect.stack()[root_idx].filename)
        return caller
    
    @staticmethod
    def determine_version(target_name: str, identify_src: bool = False) -> Union[str, Tuple[str, str]]:
        """
        Retrieve project version for distribution (or running codebase)

        Version is determined by:  
        - look for pyproject.toml  
        - check importlib metadata
        - scanning calling stack root file newest python module  

        Args:
            distrib_name (str): Package distribution name.
              If distrib_name not found, version will be determined from pyproject.toml (if found) or
              from the newest .py file in the stack path starting at the calling program.
            identify_src (bool, otional): Return a string indicating how the version was determined. Defaults to False

        Returns:
            Union[str,Tuple[str,str]: version or version, source.
            - version is in format major.minor.patch or YYYY.MM.DD.
            - source is one of 'importlib.metadata', 'pyproject.toml' or source filename.

        Raises:
            ValueError: if target_name is not str
            
        """
        if not isinstance(target_name, str):
            raise ValueError(f'Invalid target name (must be str) in determine_version: {target_name}')
        
        LOGGER.trace(f'determine_version for {target_name}')
        ver = None
        # root_idx = len(inspect.stack()) - 1
        # caller = pathlib.Path(inspect.stack()[root_idx].filename)
        caller = ProjectHelper._get_caller()
        determined_from = None
        ver, determined_from = ProjectHelper._check_toml(target_name, caller)
        if ver is None:
            ver, determined_from = ProjectHelper._check_metadata(target_name)

        if ver is None:
            python_file = f'{target_name}.py'
            ver, determined_from = ProjectHelper._check_call_stack(root_path=caller.parent, python_file=python_file)
            if ver is None:
                root_idx = 0 # len(inspect.stack()) - 2
                caller = pathlib.Path(inspect.stack()[root_idx].filename)
                ver, determined_from = ProjectHelper._check_call_stack(root_path=caller.parent, python_file=python_file)

        if identify_src:
            return (ver, determined_from)
        return ver

    @staticmethod
    def _pyproject_toml_location() -> pathlib.Path:
        caller = ProjectHelper._get_caller()
        target_file = ProjectHelper._search_down_tree("pyproject.toml", pathlib.Path(caller).parent)
        if target_file is None:
            LOGGER.trace(f'Unable to locate pyproject.toml for {caller}')
        return target_file
    
    @staticmethod
    def installed_pyproject_toml_packages() -> dict:
        """
        Return a dictionary describing pyproject.toml installed packages.

        Returns:
            dict: in format {"pkg_name1": "pkg_version", "pkg_name2": "pkg_version", ...}
        """
        toml_packages = {}
        toml_file = ProjectHelper._pyproject_toml_location()
        if toml_file is not None:
            config = tomllib.loads(toml_file.read_text())
            toml_deps = config.get('tool',{}).get('poetry', None)
            if toml_deps is not None:
                dist_deps = ProjectHelper.installed_distribution_packages()
                for toml_pkg, val in toml_deps['dependencies'].items():
                    dep_version = dist_deps.get(toml_pkg, None)
                    if dep_version is not None:
                        toml_packages[toml_pkg] = dep_version
        return toml_packages
    

    @staticmethod
    def installed_distribution_packages() -> dict:
        """
        Return a dictionary describing installed packages.

        Returns:
            dict: in format {"pkg_name1": "pkg_version", "pkg_name2": "pkg_version", ...}
        """
        package: Distribution = None
        package_dict: dict = {}
        for package in distributions():
            package_dict[package.name] = package.version
        
        return package_dict
    

if __name__ == '__main__':
    print(ProjectHelper.installed_pyproject_toml_packages())
    # print(ProjectHelper.installed_distribution_packages())


"""
Helper for OS functions.

Supports windows and linux

"""
import ctypes
import os
import pathlib
import platform
import signal
import sys
from datetime import datetime as dt
from typing import List, Tuple
import tempfile

import psutil
from loguru import logger as LOGGER

from dt_tools.misc.helpers import ObjectHelper as ohelper


class OSHelper():
    """
    Helper class for OS functions.
    
    Supports windows and linux

    Features:
        - OS/hardware detection (Windows/Linux/RaspberryPI)
        - Process detection (running in fg/bg)
        - Is executable in the path?
        - Is process running with admin/root permisssions?
        
        Windowws:
        - is_admin
        - elevate to admin

        Linux:
        - is_root

    Raises:
        OSError: On un-supported OS

    Examples::
        from dt_tools.os.os_helper import OSHelper

        print(f'Is Windows: {OSHelper.is_windows()})
        print(f'Is Linux  : {OSHelper.is_linux()})
        print(f'Is RPi    : {OSHelper.is_raspberrypi()})
        
    """
    @staticmethod
    def is_windows() -> bool:
        """Return True if running in windows else False"""
        return platform.system() == "Windows"
    
    @staticmethod
    def is_linux() -> bool:
        """Return True if running in linux else False"""
        return platform.system() == "Linux"

    @staticmethod
    def os_version() -> str:
        return f'{platform.system()} {platform.version()}'
    
    @staticmethod
    def current_user() -> str:
        return os.getlogin()
    
    @staticmethod
    def is_running_in_foreground():
        """
        Check if process is running in foreground

        Returns:
            True if running in foreground else False
        """
        # try:
        #     if os.getpgrp() == os.tcgetpgrp(sys.stdout.fileno()):
        #         return True     # is foreground
        #     return False        # is background
        # except AttributeError:
        #     # Fall back, looks like os.getpgrp() is not available
        #     return sys.stdout.isatty()
        # except OSError:
        #     return True         # is as a daemon       
        if OSHelper.is_linux():
            current_process = psutil.Process()
            return current_process.terminal() is not None
        elif OSHelper.is_windows():
            pid = os.getpid()
            for svc in psutil.win_service_iter():
                if svc.pid == pid and svc.status == 'running':
                    return False
            return True
        # Not supported OS
        LOGGER.warning('Unable to determine if process is background/foreground - Unsupported OS')
        return False
        

    @staticmethod
    def is_running_in_background():
        """
        Check if process is running in background (or as a daemon)

        Returns:
            True if running in background else False
        """
        # return not cls.is_running_in_foreground()
        return not OSHelper.is_running_in_foreground()

    @staticmethod
    def find_file(filenm: str, search_path: str = None) -> str:
        """
        Find file starting as specified path.

        Args:
            filenm (str): Filename to search for
            search_path (str): Starting search path, if None, current directory will be used.

        Returns:
            str: FQDN of file location, or None if not found.
        """
        target_fqdn: str = None
        for filepath in pathlib.Path(search_path).rglob(filenm):
            target_fqdn = filepath
            break

        return target_fqdn
            
    @staticmethod
    def is_executable_available(name: str) -> str:
        """
        Is executable in system path?

        Arguments:
            name: Name of executable.

        Returns:
            Fully qualified executable path if found, else None
        """
        if OSHelper.is_windows():
            sep = ';'
        else:
            sep = ':'
        PATH = os.getenv('PATH')
        exe = None
        found = False
        for dir in PATH.split(sep):
            exe = pathlib.Path(dir) / name
            if exe.exists():
                found = True
                break
            if OSHelper.is_windows():
                exe = pathlib.Path(dir) / f'{name}.exe'
                if exe.exists():
                    found = True
                    break
                exe = pathlib.Path(dir) / f'{name}.com'
                if exe.exists():
                    found = True
                    break

        if found:
            return exe
        return None

    @staticmethod
    def is_windows_admin():
        """
        Is process running as Windows Admin

        Returns:
            True if Admin privileges in effect else False
        """
        if OSHelper.is_windows():
            try:
                return ctypes.windll.shell32.IsUserAnAdmin()
            except Exception as ex:
                LOGGER.warning(f'On Windows, but cant check Admin privileges: {repr(ex)}')
                return False            
        
        return False

    @staticmethod
    def is_linux_root():
        """
        Is process running as root?

        Returns:
            True if root else False
        """
        return os.geteuid() == 0
    
    @staticmethod
    def is_god():
        """
        Is process running elevated.

        For windows: admin permissions
        For linux:   root user
        
        Returns:
            True if admin/root else False
        """
        if OSHelper.is_windows:
            return OSHelper.is_windows_admin()
        return OSHelper.is_linux_root()
    
    @staticmethod
    def elevate_to_admin(executable: str = None, args: List[str] = None) -> bool:
        """
        Relaunch process with elevated privileges to Windows Admin.

        User will be presented with a prompt which must be ACK'd for elevation.

        Raises:
            OSError: If not running on Windows.

        Returns:
            bool: True if successful else False
        """
        if not OSHelper.is_windows():
            raise OSError('run_as_admin is ONLY available in Windows')

        if OSHelper.is_windows_admin():
            return True
        
        tgt_executable = sys.executable if executable is None else executable
        tgt_args = sys.argv if args is None else args

        # Re-run the program with admin rights
        LOGGER.debug(f'Run Elevated - executable: {tgt_executable}   args: {tgt_args}')
        hresult = ctypes.windll.shell32.ShellExecuteW(None, "runas", tgt_executable, " ".join(tgt_args), None, 1)
        LOGGER.debug(f'  returns {hresult}')
        return True if hresult > 32 else False

    @staticmethod
    def get_temp_filename(prefix: str = None, dotted_suffix: str = None, target_dir: str = None, keep: bool = False) -> str:
        """
        Create a temporary filename

        Args:
            prefix (str, optional): Prefix filename with this string. Defaults to None.
            dotted_suffix (str, optional): Filename extension. Defaults to None.
            target_dir (str, optional): Directory for file. Defaults to None.
            keep (bool, optional): _description_. Defaults to False.

        Returns:
            str: _description_
        """
        with tempfile.NamedTemporaryFile(mode='w+b', 
                                             prefix=prefix, suffix=dotted_suffix, dir=target_dir,
                                             delete=(not keep)) as t_file:
        
            temp_filename = t_file.name

        LOGGER.debug(f'temp filename created: {temp_filename}.  Keep={keep}')
        return temp_filename
    
    # == Hardware info =============================================================================================
    @staticmethod
    def is_raspberrypi() -> bool:
        """
        Check if hardware is a Raspberry PI

        Returns:
            True if Raspberry PI else False
        """
        if not OSHelper.is_linux():
            return False
        buffer = []
        with open('/proc/cpuinfo','r') as fh:
            buffer = fh.readlines()
        token = [x for x in buffer if x.startswith('Hardware')]
        hw = token[0].split(":")[1].strip()
        if hw.startswith("BCM"):
            return True
        return False
    
    @staticmethod
    def sysinfo(include_all: bool = False, include_cpu: bool = False, include_memory: bool = False, include_disk: bool = False) -> dict:
        # TODO:  gather stats
        # system    : name, ip, processor, OS, OSVersion, Manufacturer, uptime
        # cpu       : cores
        # memory    : Space (total, used, free), type
        # disk      : Space (Total,Used,Free), disks[]
        # io
        info = {}
        info['system'] = OSHelper._get_local_system_info()
        if include_all or include_cpu:
            info['cpu'] = OSHelper._get_cpu_info()
        if include_all or include_memory:
            info['memory'] = OSHelper._get_memory_info()
        if include_all or include_disk:
            info['disk'] = OSHelper._get_disk_info()

        return info
    
    @staticmethod
    def bytes_to_printformat(num_bytes: int) -> str:
        power = 2**10
        n = 0
        power_labels = {0: '', 1: 'K', 2: 'M', 3: 'G', 4: 'T'}
        while num_bytes > power:
            num_bytes /= power
            n += 1
        return f"{num_bytes:.2f} {power_labels[n]}B"        
    
    @staticmethod
    def bytes_to_kb(num_bytes: int, num_decimals: int=0) -> float:
        kb =  num_bytes / 1024
        if num_decimals >= 0:
            return float(f'{kb:.{num_decimals}f}')
        return kb
    
    @staticmethod
    def bytes_to_mb(num_bytes: int, num_decimals: int=0) -> float:
        mb =  OSHelper.bytes_to_kb(num_bytes) / 1024
        if num_decimals >= 0:
            return float(f'{mb:.{num_decimals}f}')
        return mb
    
    @staticmethod
    def bytes_to_gb(num_bytes: int, num_decimals: int=0) -> float:
        gb = OSHelper.bytes_to_mb(num_bytes) / 1024
        if num_decimals >= 0:
            return float(f'{gb:.{num_decimals}f}')
        return gb

    @staticmethod
    def bytes_to_tb(num_bytes: int, num_decimals: int=0) -> float:
        tb =  OSHelper.bytes_to_gb(num_bytes) / 1024
        if num_decimals >= 0:
            return float(f'{tb:.{num_decimals}f}')
        return tb

    @staticmethod
    def elapsed_time(start_date: dt, end_date: dt) -> Tuple[int, int, int, int]:
        """
        Convert seconds into (days, hours, mins, sec)

        Args:
            seconds (int): number of seconds to convert

        Returns:
            Tuple[int, int, int, int]: Days, Hours, Minutes, Seconds
        """
        duration = end_date - start_date if start_date < end_date else start_date - end_date
        duration_secs = duration.total_seconds()

        days    = divmod(duration_secs, 86400)        # Get days (without [0]!)
        hours   = divmod(days[1], 3600)               # Use remainder of days to calc hours
        minutes = divmod(hours[1], 60)                # Use remainder of hours to calc minutes
        seconds = divmod(minutes[1], 1)         
        return (int(days[0]), int(hours[0]), int(minutes[0]), int(seconds[0]))

    @staticmethod
    def _get_local_system_info() -> dict:
        # uname system (os), node (name), release (osversion)
        import socket
        info = {}
        info['hostname'] = platform.node()
        info['host_fqdn'] = socket.getfqdn()
        info['ip'] = socket.gethostbyname(platform.node())  # validate this works on all platforms
        info['platform'] = platform.system()
        if OSHelper.is_linux():
            osr = platform.freedesktop_os_release()
            info['os'] = osr['NAME']
            info['os_ver'] = osr['VERSION']
        else:
            info['os'] = platform.system()
            info['os_ver'] = platform.platform()
        info['os_kernel'] = platform.uname().release
        info['machine_type'] = platform.uname().machine
        bt = psutil.boot_time()
        boot_time_str = dt.fromtimestamp(bt).strftime("%Y-%m-%d %H:%M:%S")
        boot_time = dt.strptime(boot_time_str, "%Y-%m-%d %H:%M:%S")
        days, hours, minutes, seconds = OSHelper.elapsed_time(boot_time, dt.now())
        info['last_boot_time'] = boot_time_str
        if days > 0:
            info['uptime'] = f"{days}d:{hours}h:{minutes}m:{seconds}s"
        elif hours > 0:
            info['uptime'] = f"{hours}h:{minutes}m:{seconds}s"
        elif minutes > 0:
            info['uptime'] = f"{minutes}m:{seconds}s"
        else:
            info['uptime'] = f"{seconds} secs"

        # TODO: BIOS 
        #       windows -  wmic bios get xxxxx,xxxx
        #       linux~  - /sys/class/dmi/id/ (many bios files)
        #       raspi   - no bios, but lots of info here: https://www.geeksforgeeks.org/how-to-find-all-the-hardware-information-in-raspberry-pi-os/
        return info

    @staticmethod
    def _get_cpu_info() -> dict:
        info = {}
        info['processor'] = platform.processor()
        info['cores_physical'] = psutil.cpu_count(logical=False)
        info['cores_logical'] = psutil.cpu_count(logical=True)
        cpu_freq = psutil.cpu_freq()
        info['freq_min'] = cpu_freq.min
        info['freq_max'] = cpu_freq.max
        cpu_pct = psutil.cpu_times_percent(interval=1.0) # Based on 1 sec interval
        info['pct_user'] = cpu_pct.user
        info['pct_system'] = cpu_pct.system
        info['pct_idle'] = cpu_pct.idle

        return info

    @staticmethod
    def _get_memory_info() -> dict:
        info = {}
        info['swap_total'] = psutil.swap_memory().total
        info['swap_used'] = psutil.swap_memory().used
        info['swap_free'] = psutil.swap_memory().free
        info['swap_pct_used'] = psutil.swap_memory().percent

        info['virtual_total'] = psutil.virtual_memory().total
        info['virtual_used'] = psutil.virtual_memory().used
        info['virtual_free'] = psutil.virtual_memory().free
        info['virtual_pct_used'] = psutil.virtual_memory().percent
        
        return info

    @staticmethod
    def _get_disk_info() -> dict:
        info = {}
        disk_list = []
        for partition in psutil.disk_partitions():
            entry = {}
            entry['device'] = partition.device
            entry['mountpoint'] = partition.mountpoint
            entry['fstype'] = partition.fstype
            entry['mount_opts'] = partition.opts
            try:
                du = psutil.disk_usage(partition.device)
                entry['total'] = du.total
                entry['used'] = du.used
                entry['free'] = du.free
                entry['used_pct'] = du.percent
            except OSError:
                pass
            # disk_dict[partition.device] = entry
            disk_list.append(entry)
        info['partitions'] = disk_list
        
        io = psutil.disk_io_counters()
        info['io_read_bytes'] = io.read_bytes
        info['io_write_bytes'] = io.write_bytes
        info['io_read_cnt'] = io.read_count
        info['io_write_cnt'] = io.write_count
        info['io_read_ms'] = io.read_time
        info['io_write_ms'] = io.write_time
        return info
    
    @staticmethod
    def _get_io_info() -> dict:
        info = {}
        return info

    # -- ctrl-c Handler routines ===================================================================================
    @staticmethod
    def disable_ctrl_c_handler() -> bool:
        """
        Disable handler for Ctrl-C checking.

        Returns:
          True if successful, else False
        """
        success = True
        try:
            signal.signal(signal.SIGINT, signal.SIG_DFL)
        except:  # noqa: E722
            success = False
        return success

    @staticmethod
    def enable_ctrl_c_handler(handler_function: callable = None) -> bool:
        """
        Enable handler for Ctrl-C checking.
        
        If Ctrl-C occurs, and no handler function has been defined, user is prompted to continue or exit.

        Arguments:
            handler_function: Function to be called when ctrl-c is requested. (optional) 
              If supplied, the function should be defined as follows...
            
              Example::

                def handler_name(signum, frame):
                    code to execute when handler is called...  

        Returns:
            True if handler successfully enabled, else False.

        """
        success = True
        if handler_function is None:
            handler_function = OSHelper._interrupt_handler
            
        try:
            signal.signal(signal.SIGINT, handler_function)
        except:  # noqa: E722
            success = False
        return success

    @staticmethod
    def _interrupt_handler(signum, frame):
        resp = ''
        while resp not in ['c', 'e']:
            try:
                resp = input('\nCtrl-C, Continue or Exit (c,e)? ').lower()
            except RuntimeError:
                LOGGER.error('\nCtrl-C, program exiting...')
                resp = 'e'

            if resp == 'e':
                os._exit(1)

if __name__ == "__main__":
    import json
    print(f'is foreground: {OSHelper.is_running_in_foreground()}')    
    # print(json.dumps(OSHelper.sysinfo(include_cpu=False, include_disk=True, include_memory=False), indent=2))
    # info = OSHelper.sysinfo(include_disk=True)
    # info_obj = ohelper.dict_to_obj(info)
    # print(info_obj.disk)
    # print('Device       Type            Total      Used       Free    % Used')
    # print('------------ ---------- ---------- ---------- ---------- --------')
    # for de in info_obj.disk.partitions:
    #     # print(de)
    #     total = '-'
    #     used = '-'
    #     free = '-'
    #     used_pct = '-'
    #     type = de.fstype if len(de.fstype) > 0 else de.mount_opts
    #     if hasattr(de, 'total'):
    #         total = OSHelper.bytes_to_printformat(de.total)
    #         used = OSHelper.bytes_to_printformat(de.used)
    #         free = OSHelper.bytes_to_printformat(de.free)
    #         used_pct = f'{de.used_pct}%'
    #     print(f'{de.device:12} {type:10} {total:>10} {used:>10} {free:>10} {used_pct:>8}')
    # print(OSHelper.bytes_to_kb(num_bytes, 2))
    # print(OSHelper.bytes_to_mb(num_bytes, 3))
    # print(OSHelper.bytes_to_gb(num_bytes, 1))
    # print(OSHelper.bytes_to_tb(num_bytes, 0))
    # module.demo()